// Filename: ovm_pcs_example.sv

//----------------------------------------------------------------------
//  Copyright (c) 2010 by Doulos Ltd.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.
//----------------------------------------------------------------------

// Source code example for Mentor Graphics Verification Academy OVM Basics Module

// Author: John Aynsley, Doulos
// Date:   21-Jan-2010

// Changes: Michal Kajan, 29-Nov-2012


`include "ovm_macros.svh"

interface dut_if();

  logic clock, reset;
  logic cmd;
  logic [7:0] addr;
  logic [7:0] data;

endinterface: dut_if


module dut(dut_if _if);

  always @(posedge _if.clock)
  begin
    ovm_pkg::ovm_top.ovm_report_info("PCS", $psprintf("DUT received cmd=%b, addr=%d, data=%d",
                                     _if.cmd, _if.addr, _if.data));
  end
  
endmodule: dut


package my_seq_library;

  import ovm_pkg::*;

  class my_transaction extends ovm_sequence_item;
  
    `ovm_object_utils(my_transaction)
  
    rand bit cmd;
    rand int addr;
    rand int data;
  
    constraint c_addr { addr >= 0; addr < 256; }
    constraint c_data { data >= 0; data < 256; }
    
    function new (string name = "");
      super.new(name);
    endfunction: new
    
    function string convert2string;
      return $psprintf("cmd=%b, addr=%0d, data=%0d", cmd, addr, data);
    endfunction: convert2string
  endclass: my_transaction


  class read_modify_write extends ovm_sequence #(my_transaction);
  
    `ovm_object_utils(read_modify_write)
    
    function new (string name = "");
      super.new(name);
    endfunction: new

    task body;
      my_transaction tx;
      int a;
      int d;

      tx = my_transaction::type_id::create("tx");
      start_item(tx);
      assert( tx.randomize() with { cmd == 0; } );
      finish_item(tx);

      a = tx.addr;      
      d = tx.data;
      ++d;

      tx = my_transaction::type_id::create("tx");
      start_item(tx);
      assert( tx.randomize() with { cmd == 1; addr == a; data == d; } );
      finish_item(tx);
    endtask: body
   
  endclass: read_modify_write
  

  class seq_of_commands extends ovm_sequence #(my_transaction);
  
    `ovm_object_utils(seq_of_commands)
    
    rand int n;
    
    constraint how_many { n inside {[4:6]}; }
    
    function new (string name = "");
      super.new(name);
    endfunction: new

    task body;
      repeat(n)
      begin
        read_modify_write seq;
        seq = read_modify_write::type_id::create("seq");
        start_item(seq);
        finish_item(seq);
      end
    endtask: body
   
  endclass: seq_of_commands
  
endpackage: my_seq_library


package my_pkg;

  import ovm_pkg::*;
  import my_seq_library::*;


  class dut_if_wrapper extends ovm_object;
  
    virtual dut_if dut_vi;
    
    function new(string name, virtual dut_if arg);
      super.new(name);
      dut_vi = arg;      
    endfunction: new
    
  endclass: dut_if_wrapper

  
  class my_sequencer extends ovm_sequencer #(my_transaction);
    
    `ovm_component_utils(my_sequencer)
  
    function new(string name, ovm_component parent);
      super.new(name, parent);
    endfunction: new

  endclass: my_sequencer


  class my_driver extends ovm_driver #(my_transaction);
  
    `ovm_component_utils(my_driver)

    virtual dut_if dut_vi;

    function new(string name, ovm_component parent);
      super.new(name, parent);
    endfunction: new
    
    function void build;
      super.build();
      begin
        ovm_object obj;
        dut_if_wrapper if_wrapper;
        get_config_object("dut_if_wrapper", obj, 0);
        assert( $cast(if_wrapper, obj) );
        dut_vi = if_wrapper.dut_vi;
      end
    endfunction : build
   
    task run;
      forever
      begin
        my_transaction tx;
        
        @(posedge dut_vi.clock);
        seq_item_port.get(tx);
        
        // Wiggle pins of DUT
        dut_vi.cmd  = tx.cmd;
        dut_vi.addr = tx.addr;
        dut_vi.data = tx.data;
      end
    endtask: run

  endclass: my_driver


  class my_monitor extends ovm_monitor;
  
    `ovm_component_utils(my_monitor)

    ovm_analysis_port #(my_transaction) aport;
    
    virtual dut_if dut_vi;

    function new(string name, ovm_component parent);
      super.new(name, parent);
    endfunction: new
    
    function void build;
      super.build();
      aport = new("aport", this);
      begin
        ovm_object obj;
        dut_if_wrapper if_wrapper;
        get_config_object("dut_if_wrapper", obj, 0);
        assert( $cast(if_wrapper, obj) );
        dut_vi = if_wrapper.dut_vi;
      end
    endfunction : build
   
    task run;
      forever
      begin
        my_transaction tx;
        
        @(posedge dut_vi.clock);
        tx = my_transaction::type_id::create("tx");
        tx.cmd  = dut_vi.cmd;
        tx.addr = dut_vi.addr;
        tx.data = dut_vi.data;
        
        aport.write(tx);
      end
    endtask: run

  endclass: my_monitor


  class my_agent extends ovm_agent;

    `ovm_component_utils(my_agent)
    
    ovm_analysis_port #(my_transaction) aport;
    
    my_sequencer my_sequencer_h;
    my_driver    my_driver_h;
    my_monitor   my_monitor_h;
    
    function new(string name, ovm_component parent);
      super.new(name, parent);
    endfunction: new
    
    function void build;
      super.build();
      aport = new("aport", this);
      my_sequencer_h = my_sequencer::type_id::create("my_sequencer_h", this);
      my_driver_h    = my_driver   ::type_id::create("my_driver_h"   , this);
      my_monitor_h   = my_monitor  ::type_id::create("my_monitor_h"  , this);
    endfunction: build
    
    function void connect;
      my_driver_h.seq_item_port.connect( my_sequencer_h.seq_item_export );
      my_monitor_h.       aport.connect( aport );
    endfunction: connect
    
  endclass: my_agent
  
  
  class my_subscriber extends ovm_subscriber #(my_transaction);
  
    `ovm_component_utils(my_subscriber)
    
    bit cmd;
    int addr;
    int data;
        
    covergroup cover_bus;
      coverpoint cmd;
      coverpoint addr { bins a[16] = {[0:255]}; }
      coverpoint data { bins d[16] = {[0:255]}; }
    endgroup: cover_bus
    
    function new(string name, ovm_component parent);
      super.new(name, parent);
      cover_bus = new;  
    endfunction: new

    function void build;
      super.build();
    endfunction: build
    
    function void write(my_transaction t);
      ovm_report_info("PCS", $psprintf("Subscriber received tx %s", t.convert2string()));
      
      cmd  = t.cmd;
      addr = t.addr;
      data = t.data;
      cover_bus.sample();
    endfunction: write

  endclass: my_subscriber
  
  
  class my_env extends ovm_env;

    `ovm_component_utils(my_env)
    
    OVM_FILE file_h;
    my_agent      my_agent_h;
    my_subscriber my_subscriber_h;
    
    function new(string name, ovm_component parent);
      super.new(name, parent);
    endfunction: new
    
    function void build;
      super.build();
      my_agent_h      = my_agent     ::type_id::create("my_agent_h", this);
      my_subscriber_h = my_subscriber::type_id::create("my_subscriber_h", this);
    endfunction: build
    
    function void connect;
      my_agent_h.aport.connect( my_subscriber_h.analysis_export );
    endfunction: connect
    
  endclass: my_env
  
  
  class my_test extends ovm_test;
  
    `ovm_component_utils(my_test)
    
    my_env my_env_h;   
   
    function new(string name, ovm_component parent);
      super.new(name, parent);
    endfunction: new
    
    function void build;
      super.build();
      my_env_h = my_env::type_id::create("my_env_h", this);
    endfunction: build
    
    task run;
      seq_of_commands seq;
      seq = seq_of_commands::type_id::create("seq");
      assert( seq.randomize() );
      seq.start( my_env_h.my_agent_h.my_sequencer_h );
      ovm_top.stop_request();
    endtask: run
    
  endclass: my_test
  
  
endpackage: my_pkg


module top;

  import ovm_pkg::*;
  import my_pkg::*;
  
  dut_if dut_if1 ();
  
  dut dut1 ( ._if(dut_if1) );

  // Clock and reset generator
  initial
  begin
    dut_if1.clock = 0;
    forever #5 dut_if1.clock = ~dut_if1.clock;
  end

  initial
  begin
    dut_if1.reset = 1;
    repeat(3) @(negedge dut_if1.clock);
    dut_if1.reset = 0;
  end

  initial
  begin: blk
    dut_if_wrapper if_wrapper = new("if_wrapper", dut_if1);
    set_config_object("*", "dut_if_wrapper", if_wrapper, 0);
    
    ovm_top.enable_print_topology = 1;
    ovm_top.finish_on_completion  = 1;
    ovm_top.phase_timeout         = 1ms;
    ovm_top.stop_timeout          = 1ms;
    
    run_test("my_test");
  end

endmodule: top
